//
//								apitch.cpp
//

#include <stdio.h>
#include <string.h>
#define _USE_MATH_DEFINES
#include <math.h>

#ifndef __apitch__
#include "apitch.h"
#endif

#ifndef __sdeditor__
#include "editor/sdeditor.h"
#endif

//
//								APitchProgram 
//
APitchProgram::APitchProgram ()
{
	// default Program Values
	paramFinePitch = 0.5f;
	paramCoarsePitch = 0.48f;
	paramFeedback = 0.0f;
	paramDelay = 0.0f;
	paramMixMode = 0.0f;
	strcpy (name, "Thing");
}

//
//								APitchProgram
//
APitchProgram::APitchProgram( float fine, float coarse, float feedback, float delay, float mixMode, char *pname)
{
	paramFinePitch = fine;
	paramCoarsePitch = coarse;
	paramFeedback = feedback;
	paramDelay = delay;
	paramMixMode = mixMode;
	strcpy( name, pname);
};


//
//								APitch 
//
APitch::APitch(audioMasterCallback audioMaster)
	: AudioEffectX (audioMaster, kNumPrograms, kNumParams)
{
	OutputDebugString("\n\nAPitch V0.3\n\n");

	// init programs
	pPrograms[0] = new APitchProgram( 0.515f, 0.48f, 0.0f, 0.1f, 0.0f, "Harmonizer");
	pPrograms[1] = new APitchProgram( 0.5f, 0.0f, 0.0f, 0.1f, 0.0f, "Octave Up");
	pPrograms[2] = new APitchProgram( 0.5f, 0.96f, 0.0f, 0.1f, 0.0f, "Octave Down");

	pPrograms[3] = new APitchProgram( 0.53f, 0.48f, 0.0f, 0.1f, 0.0f, "Deep Harmonizer");
	pPrograms[4] = new APitchProgram( 0.5f, 0.2f, 0.0f, 0.1f, 0.0f, "Fifth Up");
	pPrograms[5] = new APitchProgram( 0.5f, 0.76f, 0.0f, 0.1f, 0.0f, "Fifth Down");
	pPrograms[6] = new APitchProgram( 0.515f, 0.48f, 0.0f, 0.1f, 0.857f, "Stereo Harmonizer");
	pPrograms[7] = new APitchProgram( 0.5f, 0.52f, 0.3f, 0.9f, 0.0f, "Descending Echoes");

	// make sure all instance vars are init'd to some known value
	_paramFinePitch = 0.5f;
	_paramCoarsePitch = 0.48f;
	_paramFeedback = 0.0f;
	_paramDelay = 0.0f;
	_paramMixMode = 0.0f;
	_sweepRate = 0.2;
	_feedback = 0.0;
	_feedbackPhase = 1.0;
	_sweepSamples = 0;
	_delaySamples = 22;
	_sweepInc = 0.0;
	setSweep();
	_mixMode = 0;
	_fp = 0;

	_outval = 0.0f;
	_mixLeftWet =
	_mixLeftDry =
	_mixRightWet =
	_mixRightDry = 0.5f;

	// set initial program (defaults)
	setProgram (0);

	setNumInputs (2);	// pseudo-mono input
	setNumOutputs (2);	// pseudo-mono output

	setUniqueID ('qr7e');	// this should be unique, use the Steinberg web page for plugin Id registration

	// create the editor
	editor = new SDEditor (this);

	// allocate the buffer
	_buf = new double[BSZ];
	for( int i = 0; i < BSZ; i++)
	{
		_buf[i] = 0.0;
	}

	resume ();		// flush buffer
}

//
//								~APitch 
//
APitch::~APitch ()
{
	for( int i = 0; i < kNumPrograms; i++)
	{
		APitchProgram *app = pPrograms[i];
		delete app;
		pPrograms[i] = NULL;
	}
	if( _buf )
		delete[] _buf;
}

//
//								setProgram
//
void APitch::setProgram (VstInt32 program)
{
	APitchProgram* ap = pPrograms[program];

	char buf[256];
	sprintf( buf, "%d: %s\n", program, ap->name);
	OutputDebugString(buf );

	curProgram = program;
	setParameter (kFinePitch, ap->paramFinePitch);	
	setParameter (kCoarsePitch, ap->paramCoarsePitch);
	setParameter (kFeedback, ap->paramFeedback);
	setParameter (kDelay, ap->paramDelay);
	setParameter (kMixMode, ap->paramMixMode);
}

//
//								setFine
//
//	Sets the fine pitch shift amount
//
void APitch::setFine(float finePitch)
{
	_paramFinePitch = finePitch;
	pPrograms[curProgram]->paramFinePitch = finePitch;

	// finish setup
	setSweep();
}


//
//								setCoarse
//
//	Maps 0.0-1.0 input to calculated width in samples from 0ms to 10ms
//
void APitch::setCoarse (float v)
{
	_paramCoarsePitch = v;
	pPrograms[curProgram]->paramCoarsePitch = v;

	// finish setup
	setSweep();
}

//
//								setDelay
//
//	Map to the expect 0-100ms range.
//
void APitch::setDelay (float v)
{
	_paramDelay = v;
	pPrograms[curProgram]->paramDelay = v;

	// convert incoming float to int sample count for up to 100ms of delay
	_delaySamples = 0.1 * SAMPLE_RATE * v;

	// pin to a minimum value so our sweep never collides with the filling pointer
	_delaySamples = _delaySamples < MIN_DELAY_SAMPLES ? MIN_DELAY_SAMPLES : _delaySamples;

	// finish setup
	setSweep();
}

//
//								setSweep
//
//	Sets up sweep based on pitch delta and delay as they're interrelated
//	Assumes _paramFinePitch, _paramCoarsePitch, and _delaySamples have all been set by
//	setFine, setCoarse, and setDelay
//
void APitch::setSweep()
{
	// calc the total pitch delta
	double semiTones = 12.0 - ROUND(_paramCoarsePitch * NUM_PITCHES);
	double fine = (2.0 * _paramFinePitch) - 1.0;
	double pitchDelta = pow(SEMITONE,semiTones + fine);

	// see if we're increasing or decreasing
	_increasing = pitchDelta >= 1.0;

	// calc the # of samples in the sweep, 15ms minimum, scaled up to 50ms for an octave up,
	// and 32.5ms for an octave down
	double absDelta = abs(pitchDelta - 1.0);
	_sweepSamples = (MIN_SWEEP + (MAX_SWEEP-MIN_SWEEP) * absDelta) * SAMPLE_RATE;

	// fix up the pitchDelta to become the _sweepInc
	_sweepInc = pitchDelta - 1.0;
	_sweepInc = -_sweepInc;

	// assign initial pointers
	_sweepB = _sweepSamples / 2.0;
	_sweepA = _increasing ? _sweepSamples : 0.0;
}


//
//								setFeedback
//
void APitch::setFeedback(float v)
{
	_paramFeedback = v;
	pPrograms[curProgram]->paramFeedback = v;
	_feedback = v;
}


//
//								setMixMode
//
//	Expects input 0.0, 0.2, 0.4, 0.6, 0.8 and maps to the five supported mix modes
//
void APitch::setMixMode (float v)
{
	_paramMixMode = v;
	pPrograms[curProgram]->paramMixMode = v;
	_mixMode = (int)ROUND(v * NUM_MIX_MODES);
	switch(_mixMode)
	{
	case kMixMono:
	default:
		_mixLeftWet = _mixRightWet = 1.0f;
		_mixLeftDry = _mixRightDry = 1.0f;
		_feedbackPhase = 1.0;
		break;
	case kMixWetOnly:
		_mixLeftWet = _mixRightWet = 1.0f;
		_mixLeftDry = _mixRightDry = 0.0f;
		_feedbackPhase = 1.0;
		break;
	case kMixWetLeft:
		_mixLeftWet = _mixRightDry = 1.0;
		_mixRightWet = _mixLeftDry = 0.0;
		_feedbackPhase = 1.0;
		break;
	case kMixWetRight:
		_mixLeftWet = _mixRightDry = 0.0;
		_mixRightWet = _mixLeftDry = 1.0;
		_feedbackPhase = 1.0;
		break;
	case kMixWetLeftish:
		_mixLeftWet = _mixRightDry = 0.87;
		_mixRightWet = _mixLeftDry = 0.13;
		_feedbackPhase = 1.0;
		break;
	case kMixWetRightish:
		_mixLeftWet = _mixRightDry = 0.13;
		_mixRightWet = _mixLeftDry = 0.87;
		_feedbackPhase = 1.0;
		break;
	case kMixStereo:
		_mixLeftWet = 1.0f;
		_mixLeftDry = 1.0f;
		_mixRightWet = -1.0f;
		_mixRightDry = 1.0f;
		break;
	}
}



//
//								setProgramName
//
void APitch::setProgramName (char *name)
{
	strcpy (pPrograms[curProgram]->name, name);
}

//
//								getProgramName
//
void APitch::getProgramName (char *name)
{
	if (!strcmp (pPrograms[curProgram]->name, "Thing"))
		sprintf (name, "%s %d", pPrograms[curProgram]->name, curProgram + 1);
	else
		strcpy (name, pPrograms[curProgram]->name);
}

//
//								getProgramNameIndexed
//
bool APitch::getProgramNameIndexed (VstInt32 category, VstInt32 index, char* text)
{
	if (index < kNumPrograms)
	{
		strcpy (text, pPrograms[index]->name);
		return true;
	}
	return false;
}

//
//								resume
//
void APitch::resume ()
{
	AudioEffectX::resume ();
}

//
//								setParameter
//
void APitch::setParameter (VstInt32 index, float value)
{
	APitchProgram* ap = pPrograms[curProgram];

	switch (index)
	{
		case kFinePitch:    setFine(value);				break;
		case kCoarsePitch:   setCoarse(value);			break;
		case kFeedback: setFeedback(value);				break;
		case kDelay:	setDelay(value);				break;
		case kMixMode: setMixMode(value);				break;
	}

	if( editor != NULL)
	{
		SDEditor *sde = (SDEditor *)editor;
		sde->setParameter(index, value);
	}
}

//
//								getParameter
//
float APitch::getParameter (VstInt32 index)
{
	float v = 0;

	switch (index)
	{
		case kFinePitch:    v = _paramFinePitch;	break;
		case kCoarsePitch:	v = _paramCoarsePitch;	break;
		case kFeedback: v = _paramFeedback;			break;
		case kDelay:	v = _paramDelay;			break;
		case kMixMode:	v = _paramMixMode;			break;
	}
	return v;
}

//
//								getParameterName
//
void APitch::getParameterName (VstInt32 index, char *label)
{
	switch (index)
	{
		case kFinePitch:    strcpy (label, "Fine Pitch");			break;
		case kCoarsePitch:   strcpy (label, "Coarse Pitch");		break;
		case kFeedback: strcpy(label, "Feedback");					break;
		case kDelay:	strcpy (label, "Delay");					break;
		case kMixMode: strcpy (label, "Mix Mode");					break;
	}
}

//
//								getParameterDisplay
//
void APitch::getParameterDisplay (VstInt32 index, char *text)
{
	char buf[64];
	switch (index)
	{
	case kFinePitch:    
		sprintf( buf, "%f", _paramFinePitch);
		strcpy( text, buf );
		break;
	case kCoarsePitch:
		sprintf( buf, "%f", _paramCoarsePitch);
		strcpy( text, buf );
		break;
	case kFeedback:
		sprintf( buf, "%2.0f", _feedback * 100.0 );
		strcpy( text,buf );
		break;
	case kDelay:
		sprintf( buf, "%d", _paramDelay);
		break;
	case kMixMode:
		switch(_mixMode)
		{
		case kMixMono:
			strcpy( text, "mono" );
			break;
		case kMixWetOnly:
			strcpy( text, "mono wet only" );
			break;
		case kMixWetLeft:
			strcpy( text, "wet left" );
			break;
		case kMixWetRight:
			strcpy( text, "wet right" );
			break;
		case kMixWetLeftish:
			strcpy( text, "wet part left" );
			break;
		case kMixWetRightish:
			strcpy( text, "wet part right" );
			break;
		case kMixStereo:
			strcpy( text, "stereo" );
			break;
		}
		break;
	}
}

//
//								getParameterLabel
//
void APitch::getParameterLabel (VstInt32 index, char *label)
{
	switch (index)
	{
		case kFinePitch:		strcpy (label, "");			break;
		case kCoarsePitch:	strcpy( label, "semi-tones");	break;
		case kFeedback:	strcpy( label, "%");				break;
		case kDelay:	strcpy( label, "");					break;
		case kMixMode:	strcpy( label, "");					break;
	}
}

//
//								getEffectName
//
bool APitch::getEffectName (char* name)
{
	strcpy (name, "Pitch");
	return true;
}

//
//								getProductString
//
bool APitch::getProductString (char* text)
{
	strcpy (text, "Pitch");
	return true;
}

//
//								getVendorString
//
bool APitch::getVendorString (char* text)
{
	strcpy (text, "DarkVapor");
	return true;
}

//
//								processReplacing
//
void APitch::processReplacing (float** inputs, float** outputs, VstInt32 sampleFrames)
{
	float* in1 = inputs[0];
	float* in2 = inputs[1];
	float* out1 = outputs[0];
	float* out2 = outputs[1];

	// iterate thru sample frames
	for( int i = 0; i < sampleFrames; i++ )
	{
		// assemble mono input value and store it in circle queue
		float inval = (in1[i] + in2[i]) / 2.0f;
		double inmix = inval + _feedback * _feedbackPhase * _outval;

		_buf[_fp] = inmix;
		_fp = (_fp + 1) & (BSZ-1);

		// do the two taps
		_outval = 0.0;

		// channel A build the two emptying pointers and do linear interpolation
		int ep1, ep2;
		double w1, w2;
		double ep = _fp - _delaySamples;
		ep -= _sweepA;
		if( ep < BSZ )
		{
			ep += BSZ;
		}
		MODF(ep, ep1, w2);
		ep1 &= (BSZ-1);
		ep2 = ep1 + 1;
		ep2 &= (BSZ-1);
		w1 = 1.0 - w2;
		double tapout = _buf[ep1] * w1 + _buf[ep2] * w2;
		double fade = sin((_sweepA / _sweepSamples) * M_PI);
		tapout *= fade;
		_outval += tapout;

		// step the sweep
		_sweepA += _sweepInc;
		if( _sweepA < 0 )
		{
			_sweepA = _sweepSamples;
		}
		else if( _sweepA >= _sweepSamples)
		{
			_sweepA = 0.0;
		}

		// channel B build the two emptying pointers and do linear interpolation
		ep = _fp - _delaySamples;
		ep -= _sweepB;
		if( ep < BSZ )
		{
			ep += BSZ;
		}
		MODF(ep, ep1, w2);
		ep1 &= (BSZ-1);
		ep2 = ep1 + 1;
		ep2 &= (BSZ-1);
		w1 = 1.0 - w2;
		tapout = _buf[ep1] * w1 + _buf[ep2] * w2;
		fade = sin((_sweepB / _sweepSamples) * M_PI);
		tapout *= fade;
		_outval += tapout;

		// step the sweep
		_sweepB += _sweepInc;
		if( _sweepB < 0 )
		{
			_sweepB = _sweepSamples;
		}
		else if( _sweepB >= _sweepSamples)
		{
			_sweepB = 0.0;
		}
		
		// develop output mix 
		out1[i] = (float)PIN(_mixLeftDry * inval + _mixLeftWet * _outval,-0.99,0.99);
		out2[i] = (float)PIN(_mixRightDry * inval + _mixRightWet * _outval,-0.99,0.99);

	} // end of loop iterating thru sample frames
}
